from mlregistry.registry import register as register
from mlregistry.registry import Registry as Registry
from mlregistry.accessors import getarguments as getarguments
from mlregistry.accessors import gethash as gethash
from mlregistry.accessors import getname as getname
from mlregistry.accessors import sethash as sethash
from mlregistry.accessors import setname as setname